@@ -7,14 +7,20 @@ from functools import wraps
 
 from astropy import units as u
 from astropy.extern import six
-from astropy.io.fits import PrimaryHDU, ImageHDU
+from astropy.io.fits import PrimaryHDU, ImageHDU, Header, Card
+from astropy import log
+from astropy import wcs
 
 import numpy as np
 
 from . import cube_utils
 from . import wcs_utils
-from .masks import LazyMask, BooleanArrayMask
+from . import spectral_axis
+from .masks import LazyMask, BooleanArrayMask, MaskBase, is_broadcastable
 from .io.core import determine_format
+from .ytcube import ytCube
+
+from distutils.version import StrictVersion
 
 __all__ = ['SpectralCube']
 
@@ -23,6 +29,12 @@ try:  # TODO replace with six.py
 except NameError:
     xrange = range
 
+try:
+    from scipy import ndimage
+    scipyOK = True
+except ImportError:
+    scipyOK = False
+
 
 DOPPLER_CONVENTIONS = {}
 DOPPLER_CONVENTIONS['radio'] = u.doppler_radio
@@ -75,7 +87,8 @@ np2wcs = {2: 0, 1: 1, 0: 2}
 
 class Projection(u.Quantity):
 
-    def __new__(cls, value, unit=None, dtype=None, copy=True, wcs=None, meta=None):
+    def __new__(cls, value, unit=None, dtype=None, copy=True, wcs=None,
+                meta=None, mask=None):
 
         if value.ndim != 2:
             raise ValueError("value should be a 2-d array")
@@ -86,6 +99,7 @@ class Projection(u.Quantity):
         self = u.Quantity.__new__(cls, value, unit=unit, dtype=dtype, copy=copy).view(cls)
         self._wcs = wcs
         self._meta = meta
+        self._mask = mask
 
         return self
 
@@ -128,15 +142,36 @@ class Projection(u.Quantity):
             raise ValueError("Unknown format '{0}' - the only available "
                              "format at this time is 'fits'")
 
+    def quicklook(self):
+        """
+        Use aplpy to make a quick-look image of the projection.  This will make
+        the `FITSFigure` attribute available.
+        """
+        if not hasattr(self, 'FITSFigure'):
+            import aplpy
+            self.FITSFigure = aplpy.FITSFigure(self.hdu)
+
+        self.FITSFigure.show_grayscale()
+        self.FITSFigure.add_colorbar()
+
+# A slice is just like a projection in every way
+class Slice(Projection):
+    pass
+
 
 class SpectralCube(object):
 
-    def __init__(self, data, wcs, mask=None, meta=None, fill_value=np.nan):
+    def __init__(self, data, wcs, mask=None, meta=None, fill_value=np.nan,
+                 header=None):
 
         # Deal with metadata first because it can affect data reading
         self._meta = meta or {}
         if 'BUNIT' in self._meta:
-            self._unit = u.Unit(self._meta['BUNIT'])
+            try:
+                self._unit = u.Unit(self._meta['BUNIT'])
+            except ValueError:
+                warnings.warn("Could not parse unit {0}".format(self._meta['BUNIT']))
+                self._unit = None
         elif hasattr(data, 'unit'):
             self._unit = data.unit
             # strip the unit so that it can be treated as cube metadata
@@ -157,7 +192,12 @@ class SpectralCube(object):
         # However, we do want to keep track of what units the spectral axis
         # should be returned in, otherwise astropy's WCS can change the units,
         # e.g. km/s -> m/s.
-        self._spectral_unit = None
+        self._spectral_unit = u.Unit(self._wcs.wcs.cunit[2])
+        self._spectral_scale = 1.0
+
+        self._header = Header() if header is None else header
+        if not isinstance(self._header, Header):
+            raise TypeError("If a header is given, it must be a fits.Header")
 
     def _new_cube_with(self, data=None, wcs=None, mask=None, meta=None,
                        fill_value=None, spectral_unit=None):
@@ -172,6 +212,8 @@ class SpectralCube(object):
         cube = SpectralCube(data=data, wcs=wcs, mask=mask, meta=meta,
                             fill_value=fill_value)
         cube._spectral_unit = spectral_unit
+        cube._spectral_scale = spectral_axis.wcs_unit_scale(spectral_unit)
+        cube._header = self._header
 
         return cube
 
@@ -209,25 +251,70 @@ class SpectralCube(object):
         s += " n_s: {0}  type_s: {1:8s}  unit_s: {2}".format(self.shape[0], self.wcs.wcs.ctype[2], self.wcs.wcs.cunit[2])
         return s
 
-    def _apply_numpy_function(self, function, fill=np.nan,
-                              reduce=True, how='auto',
-                              check_endian=False, **kwargs):
+    def apply_numpy_function(self, function, fill=np.nan,
+                             reduce=True, how='auto',
+                             projection=False,
+                             unit=None,
+                             check_endian=False, **kwargs):
         """
         Apply a numpy function to the cube
+
+        Parameters
+        ----------
+        function : `numpy.ufunc`
+            A numpy ufunc to apply to the cube
+        fill : float
+            The fill value to use on the data
+        reduce : bool
+            reduce indicates whether this is a reduce-like operation,
+            that can be accumulated one slice at a time.
+            sum/max/min are like this. argmax/argmin are not
+        how : cube | slice | ray | auto
+           How to compute the moment. All strategies give the same
+           result, but certain strategies are more efficient depending
+           on data size and layout. Cube/slice/ray iterate over
+           decreasing subsets of the data, to conserve memory.
+           Default='auto'
+        projection : bool
+            Return a `Projection` if the resulting array is 2D?
+        unit : None or `astropy.units.Unit`
+            The unit to include for the output array.  For example,
+            `SpectralCube.max` calls `SpectralCube.apply_numpy_function(np.max,
+            unit=self.unit)`, inheriting the unit from the original cube.
+            However, for other numpy functions, e.g. `numpy.argmax`, the return
+            is an index and therefore unitless.
+        check_endian : bool
+            A flag to check the endianness of the data before applying the
+            function.  This is only needed for optimized functions, e.g. those
+            in the `bottleneck` package.
+        kwargs : dict
+            Passed to the numpy function.
+
+        Returns
+        -------
+        result : `Projection` or `~astropy.units.Quantity` or float
+            The result depends on the value of ``axis``, ``projection``, and
+            ``unit``.  If ``axis`` is None, the return will be a scalar with or
+            without units.  If axis is an integer, the return will be a
+            `Projection` if ``projection`` is set
         """
 
-        # reduce indicates whether this is a reduce-like operation,
-        # that can be accumulated one slice at a time.
-        # sum/max/min are like this. argmax/argmin are not
+
+        # leave axis in kwargs to avoid overriding numpy defaults, e.g.  if the
+        # default is axis=-1, we don't want to force it to be axis=None by
+        # specifying that in the function definition
+        axis = kwargs.get('axis', None)
 
         if how == 'auto':
-            strategy = cube_utils.iterator_strategy(self, kwargs.get('axis', None))
+            strategy = cube_utils.iterator_strategy(self, axis)
         else:
             strategy = how
 
+        out = None
+
         if strategy == 'slice' and reduce:
             try:
-                return self._reduce_slicewise(function, fill,
+                out = self._reduce_slicewise(function, fill,
                                               check_endian,
                                               **kwargs)
             except NotImplementedError:
@@ -236,9 +323,26 @@ class SpectralCube(object):
         if how not in ['auto', 'cube']:
             warnings.warn("Cannot use how=%s. Using how=cube" % how)
 
-        return function(self._get_filled_data(fill=fill,
-                                              check_endian=check_endian),
-                        **kwargs)
+        if out is None:
+            out = function(self._get_filled_data(fill=fill,
+                                                 check_endian=check_endian),
+                           **kwargs)
+
+        if axis is None:
+            # return is scalar
+            if unit is not None:
+                return u.Quantity(out, unit=unit)
+            else:
+                return out
+        elif projection and reduce:
+            new_wcs = wcs_utils.drop_axis(self._wcs, np2wcs[axis])
+
+            meta = {'collapse_axis': axis}
+
+            return Projection(out, copy=False, wcs=new_wcs, meta=meta, unit=unit)
+        else:
+            return out
+
 
     def _reduce_slicewise(self, function, fill, check_endian, **kwargs):
         """
@@ -272,38 +376,87 @@ class SpectralCube(object):
     @property
     def mask(self):
         """
-        The underling mask
+        The underlying mask
         """
         return self._mask
 
+    def _naxes_dropped(self, view):
+        """
+        Determine how many axes are being selected given a view.
+
+        (1,2) -> 2
+        None -> 3
+        1 -> 1
+        2 -> 1
+        """
+
+        if hasattr(view,'__len__'):
+            return len(view)
+        elif view is None:
+            return 3
+        else:
+            return 1
+
     @aggregation_docstring
     def sum(self, axis=None, how='auto'):
         """
         Return the sum of the cube, optionally over an axis.
         """
 
-        # use nansum, and multiply by mask to add zero each time there is badness
-        return u.Quantity(self._apply_numpy_function(np.nansum, fill=np.nan,
-                                                     how=how, axis=axis), self.unit,
-                          copy=False)
+        projection = self._naxes_dropped(axis) == 1
+
+        return self.apply_numpy_function(np.nansum, fill=np.nan, how=how,
+                                         axis=axis, unit=self.unit,
+                                         projection=projection)
+
+    @aggregation_docstring
+    def mean(self, axis=None, how='auto'):
+        """
+        Return the mean of the cube, optionally over an axis.
+        """
+
+        projection = self._naxes_dropped(axis) == 1
+
+        return self.apply_numpy_function(np.nanmean, fill=np.nan, how=how,
+                                         axis=axis, unit=self.unit,
+                                         projection=projection)
+
+    @aggregation_docstring
+    def std(self, axis=None, how='auto'):
+        """
+        Return the standard deviation of the cube, optionally over an axis.
+        """
+
+        projection = self._naxes_dropped(axis) == 1
+
+        return self.apply_numpy_function(np.nanstd, fill=np.nan, how=how,
+                                         axis=axis, unit=self.unit,
+                                         projection=projection)
+
 
     @aggregation_docstring
     def max(self, axis=None, how='auto'):
         """
         Return the maximum data value of the cube, optionally over an axis.
         """
-        return u.Quantity(self._apply_numpy_function(np.nanmax, fill=np.nan,
-                                                     how=how, axis=axis), self.unit,
-                          copy=False)
+
+        projection = self._naxes_dropped(axis) == 1
+
+        return self.apply_numpy_function(np.nanmax, fill=np.nan, how=how,
+                                         axis=axis, unit=self.unit,
+                                         projection=projection)
 
     @aggregation_docstring
     def min(self, axis=None, how='auto'):
         """
         Return the minimum data value of the cube, optionally over an axis.
         """
-        return u.Quantity(self._apply_numpy_function(np.nanmin, fill=np.nan,
-                                                     how=how, axis=axis), self.unit,
-                          copy=False)
+
+        projection = self._naxes_dropped(axis) == 1
+
+        return self.apply_numpy_function(np.nanmin, fill=np.nan, how=how,
+                                         axis=axis, unit=self.unit,
+                                         projection=projection)
 
     @aggregation_docstring
     def argmax(self, axis=None, how='auto'):
@@ -313,9 +466,9 @@ class SpectralCube(object):
         The return value is arbitrary if all pixels along ``axis`` are
         excluded from the mask.
         """
-        return self._apply_numpy_function(np.nanargmax, fill=-np.inf,
-                                          reduce=False,
-                                          how=how, axis=axis)
+        return self.apply_numpy_function(np.nanargmax, fill=-np.inf,
+                                         reduce=False, projection=False,
+                                         how=how, axis=axis)
 
     @aggregation_docstring
     def argmin(self, axis=None, how='auto'):
@@ -325,9 +478,9 @@ class SpectralCube(object):
         The return value is arbitrary if all pixels along ``axis`` are
         excluded from the mask
         """
-        return self._apply_numpy_function(np.nanargmin, fill=np.inf,
-                                          reduce=False,
-                                          how=how, axis=axis)
+        return self.apply_numpy_function(np.nanargmin, fill=np.inf,
+                                         reduce=False, projection=False,
+                                         how=how, axis=axis)
 
     def chunked(self, chunksize=1000):
         """
@@ -349,26 +502,43 @@ class SpectralCube(object):
         ny = self.shape[iteraxes[1]]
         return nx, ny
 
-    def _apply_along_axes(self, function, axis=None, weights=None, wcs=False,
-                          **kwargs):
+    def apply_function(self, function, axis=None, weights=None, unit=None,
+                       projection=False, **kwargs):
         """
-        Apply a function to valid data along the specified axis, optionally
-        using a weight array that is the same shape (or at least can be sliced
-        in the same way)
+        Apply a function to valid data along the specified axis or to the whole
+        cube, optionally using a weight array that is the same shape (or at
+        least can be sliced in the same way)
 
         Parameters
         ----------
-        function: function
+        function : function
             A function that can be applied to a numpy array.  Does not need to
             be nan-aware
-        axis: int
-            The axis to operate along
-        weights: (optional) np.ndarray
+        axis : 1, 2, 3, or None
+            The axis to operate along.  If None, the return is scalar.
+        weights : (optional) np.ndarray
             An array with the same shape (or slicing abilities/results) as the
             data cube
+        unit : (optional) `~astropy.units.Unit`
+            The unit of the output projection or value.  Not all functions
+            should return quantities with units.
+        projection : bool
+            Return a projection if the resulting array is 2D?
+
+        Returns
+        -------
+        result : `Projection` or `~astropy.units.Quantity` or float
+            The result depends on the value of ``axis``, ``projection``, and
+            ``unit``.  If ``axis`` is None, the return will be a scalar with or
+            without units.  If axis is an integer, the return will be a
+            `Projection` if ``projection`` is set
         """
         if axis is None:
-            return function(self.flattened(), **kwargs)
+            out = function(self.flattened(), **kwargs)
+            if unit is not None:
+                return u.Quantity(out, unit=unit)
+            else:
+                return out
 
         # determine the output array shape
         nx, ny = self._get_flat_shape(axis)
@@ -384,11 +554,14 @@ class SpectralCube(object):
                 # store result in array
                 out[x, y] = function(data, **kwargs)
 
-        if wcs:
-            newwcs = wcs_utils.drop_axis(self._wcs, np2wcs[axis])
-            return out, newwcs
+        if projection and axis in (0,1,2):
+            new_wcs = wcs_utils.drop_axis(self._wcs, np2wcs[axis])
 
-        return out
+            meta = {'collapse_axis': axis}
+
+            return Projection(out, copy=False, wcs=new_wcs, meta=meta, unit=unit)
+        else:
+            return out
 
     def _iter_rays(self, axis=None):
         """
@@ -455,14 +628,13 @@ class SpectralCube(object):
         """
         try:
             from bottleneck import nanmedian
-            return u.Quantity(self._apply_numpy_function(nanmedian, axis=axis,
-                                                         check_endian=True,
-                                                         **kwargs), self.unit,
-                              copy=False)
+            result = self.apply_numpy_function(nanmedian, axis=axis,
+                                                projection=True,
+                                                check_endian=True, **kwargs)
         except ImportError:
-            return u.Quantity(self._apply_along_axes(np.median, axis=axis,
-                                                     **kwargs), self.unit,
-                              copy=False)
+            result = self.apply_function(np.median, axis=axis, **kwargs)
+
+        return result
 
     def percentile(self, q, axis=None, **kwargs):
         """
@@ -475,9 +647,7 @@ class SpectralCube(object):
         axis : int, or None
             Which axis to compute percentiles over
         """
-        return u.Quantity(self._apply_along_axes(np.percentile, q=q, axis=axis,
-                                                 **kwargs), self.unit,
-                          copy=False)
+        return self.apply_function(np.percentile, q=q, axis=axis, **kwargs)
 
     def with_mask(self, mask, inherit_mask=True):
         """
@@ -513,13 +683,39 @@ class SpectralCube(object):
         return self._new_cube_with(mask=self._mask & mask if inherit_mask else mask)
 
     def __getitem__(self, view):
+
+        # Need to allow self[:], self[:,:]
+        if isinstance(view, slice):
+            view = (view, slice(None), slice(None))
+        elif len(view) == 2:
+            view = view + (slice(None),)
+        elif len(view) > 3:
+            raise IndexError("Too many indices")
+
         meta = {}
         meta.update(self._meta)
-        meta['slice'] = [(s.start, s.stop, s.step) for s in view]
+        meta['slice'] = [(s.start, s.stop, s.step)
+                         if hasattr(s,'start') else s
+                         for s in view]
+
+        intslices = [2-ii for ii,s in enumerate(view) if not hasattr(s,'start')]
+
+        if intslices:
+            if len(intslices) > 1:
+                # TODO: return a Specutils Spectrum object
+                raise NotImplementedError("1D slices are not implemented yet.")
+            # only one element, so drop an axis
+            newwcs = wcs_utils.drop_axis(self._wcs, intslices[0])
+            return Slice(value=self.filled_data[view],
+                         wcs=newwcs,
+                         copy=False,
+                         meta=meta)
+
+        newmask = self._mask[view] if self._mask is not None else None
 
         return self._new_cube_with(data=self._data[view],
                                    wcs=wcs_utils.slice_wcs(self._wcs, view),
-                                   mask=self._mask[view],
+                                   mask=newmask,
                                    meta=meta)
 
     @property
@@ -576,29 +772,35 @@ class SpectralCube(object):
             wavelength/frequency can be overridden with this parameter.
 
         """
-        from .spectral_axis import convert_spectral_axis,determine_ctype_from_vconv
+        from .spectral_axis import (convert_spectral_axis,
+                                    determine_ctype_from_vconv)
 
         if velocity_convention in DOPPLER_CONVENTIONS:
             velocity_convention = DOPPLER_CONVENTIONS[velocity_convention]
 
+        # Shorter versions to keep lines under 80
+        ctype_from_vconv = determine_ctype_from_vconv
+        vc = velocity_convention
+
         meta = self._meta.copy()
         if 'Original Unit' not in self._meta:
             meta['Original Unit'] = self._wcs.wcs.cunit[self._wcs.wcs.spec]
             meta['Original Type'] = self._wcs.wcs.ctype[self._wcs.wcs.spec]
 
-        out_ctype = determine_ctype_from_vconv(self._wcs.wcs.ctype[self._wcs.wcs.spec],
-                                               unit,
-                                               velocity_convention=velocity_convention)
+        out_ctype = ctype_from_vconv(self._wcs.wcs.ctype[self._wcs.wcs.spec],
+                                     unit,
+                                     velocity_convention=velocity_convention)
 
         newwcs = convert_spectral_axis(self._wcs, unit, out_ctype,
                                        rest_value=rest_value)
 
         newmask = self._mask.with_spectral_unit(unit,
-                                                velocity_convention=velocity_convention,
+                                                velocity_convention=vc,
                                                 rest_value=rest_value)
         newmask._wcs = newwcs
 
-        cube = self._new_cube_with(wcs=newwcs, mask=newmask, meta=meta, spectral_unit=unit)
+        cube = self._new_cube_with(wcs=newwcs, mask=newmask, meta=meta,
+                                   spectral_unit=unit)
 
         return cube
 
@@ -688,6 +890,7 @@ class SpectralCube(object):
         y = y.reshape(1, y.shape[0], y.shape[1])
         spectral = spectral.reshape(-1, 1, 1) - spectral.ravel()[0]
         x, y, spectral = np.broadcast_arrays(x, y, spectral)
+
         return spectral, y, x
 
     @cached
@@ -762,6 +965,9 @@ class SpectralCube(object):
         dspectral = np.abs(dspectral.reshape(-1, 1, 1))
         dx, dy, dspectral = np.broadcast_arrays(dx, dy, dspectral)
 
+        # Take spectral units into account
+        dspectral = dspectral * self._spectral_scale
+
         return dspectral, dy, dx
 
     def moment(self, order=0, axis=0, how='auto'):
@@ -830,10 +1036,16 @@ class SpectralCube(object):
 
         # apply units
         if order == 0:
-            axunit = unit = u.Unit(self._wcs.wcs.cunit[np2wcs[axis]])
+            if axis == 0 and self._spectral_unit is not None:
+                axunit = unit = self._spectral_unit
+            else:
+                axunit = unit = u.Unit(self._wcs.wcs.cunit[np2wcs[axis]])
             out = u.Quantity(out, self.unit * axunit, copy=False)
         else:
-            unit = u.Unit(self._wcs.wcs.cunit[np2wcs[axis]]) ** max(order, 1)
+            if axis == 0 and self._spectral_unit is not None:
+                unit = self._spectral_unit ** max(order, 1)
+            else:
+                unit = u.Unit(self._wcs.wcs.cunit[np2wcs[axis]]) ** max(order, 1)
             out = u.Quantity(out, unit, copy=False)
 
         # special case: for order=1, axis=0, you usually want
@@ -882,7 +1094,7 @@ class SpectralCube(object):
         """
         The `~astropy.units.equivalencies` that describes the spectral axis
         """
-        return wcs_utils.determine_vconv_from_ctype(self.wcs.wcs.ctype[self.wcs.wcs.spec])
+        return spectral_axis.determine_vconv_from_ctype(self.wcs.wcs.ctype[self.wcs.wcs.spec])
 
     @property
     def spatial_coordinate_map(self):
@@ -965,11 +1177,14 @@ class SpectralCube(object):
             try:
                 mask_slab = self._mask[ilo:ihi, :, :]
             except NotImplementedError:
-                warnings.warn("Mask slicing not implemented for {0} - dropping mask".format(self._mask.__class__.__name__))
+                warnings.warn("Mask slicing not implemented for "
+                              "{0} - dropping mask".
+                              format(self._mask.__class__.__name__))
                 mask_slab = None
 
         # Create new spectral cube
-        slab = self._new_cube_with(data=self._data[ilo:ihi], wcs=wcs_slab, mask=mask_slab)
+        slab = self._new_cube_with(data=self._data[ilo:ihi], wcs=wcs_slab,
+                                   mask=mask_slab)
 
         # TODO: we could change the WCS to give a spectral axis in the
         # correct units as requested - so if the initial cube is in Hz and we
@@ -978,16 +1193,175 @@ class SpectralCube(object):
 
         return slab
 
-    def subcube(self, xlo, xhi, ylo, yhi, zlo, zhi, rest_value=None):
+    def minimal_subcube(self):
+        """
+        Return the minimum enclosing subcube where the mask is valid
+        """
+        return self[self.subcube_slices_from_mask(self._mask)]
+
+    def subcube_from_mask(self, region_mask):
+        """
+        Given a mask, return the minimal subcube that encloses the mask
+
+        Parameters
+        ----------
+        region_mask: `masks.MaskBase` or boolean `numpy.ndarray`
+            The mask with appropraite WCS or an ndarray with matched
+            coordinates
+        """
+        return self[self.subcube_slices_from_mask(region_mask)]
+
+
+    def subcube_slices_from_mask(self, region_mask):
+        """
+        Given a mask, return the slices corresponding to the minimum subcube
+        that encloses the mask
+
+        Parameters
+        ----------
+        region_mask: `masks.MaskBase` or boolean `numpy.ndarray`
+            The mask with appropraite WCS or an ndarray with matched
+            coordinates
+        """
+        if not scipyOK:
+            raise ImportError("Scipy could not be imported: this function won't work.")
+
+        if isinstance(region_mask, np.ndarray):
+            if is_broadcastable(region_mask.shape, self.shape):
+                region_mask = BooleanArrayMask(region_mask, self._wcs)
+            else:
+                raise ValueError("Mask shape does not match cube shape.")
+
+        include = region_mask.include(self._data, self._wcs)
+
+        slices = ndimage.find_objects(np.broadcast_arrays(include,
+                                                          self._data)[0])[0]
+
+        return slices
+
+    def subcube(self, xlo='min', xhi='max', ylo='min', yhi='max', zlo='min',
+                zhi='max', rest_value=None):
         """
         Extract a sub-cube spatially and spectrally.
 
-        This method is not yet implemented.
-
-        xlo = 'min' / 'max' should be special keywords
-
+        Parameters
+        ----------
+        [xyz]lo/[xyz]hi : int or `Quantity` or `min`/`max`
+            The endpoints to extract.  If given as a quantity, will be
+            interpreted as World coordinates.  If given as a string or
+            int, will be interpreted as pixel coordinates.
         """
-        raise NotImplementedError()
+
+        limit_dict = {'xlo':0 if xlo == 'min' else xlo,
+                      'ylo':0 if ylo == 'min' else ylo,
+                      'zlo':0 if zlo == 'min' else zlo,
+                      'xhi':self.shape[2] if xhi=='max' else xhi,
+                      'yhi':self.shape[1] if yhi=='max' else yhi,
+                      'zhi':self.shape[0] if zhi=='max' else zhi}
+        dims = {'x': 2,
+                'y': 1,
+                'z': 0}
+
+        # Specific warning for slicing a frequency axis with a velocity or
+        # vice/versa
+        if ((hasattr(zlo, 'unit') and not
+             zlo.unit.is_equivalent(self.spectral_axis.unit)) or
+            (hasattr(zhi, 'unit') and not
+             zhi.unit.is_equivalent(self.spectral_axis.unit))):
+            raise u.UnitsError("Spectral units are not equivalent to the "
+                               "spectral slice.  Use `.with_spectral_unit` "
+                               "to convert to equivalent units first")
+
+        for val in (xlo,ylo,xhi,yhi):
+            if hasattr(val, 'unit') and not val.unit.is_equivalent(u.degree):
+                raise u.UnitsError("The X and Y slices must be specified in "
+                                   "degree-equivalent units.")
+
+        for lim in limit_dict:
+            limval = limit_dict[lim]
+            if hasattr(limval, 'unit'):
+                dim = dims[lim[0]]
+                sl = [slice(0,1)]*2
+                sl.insert(dim, slice(None))
+                spine = self.world[sl][dim]
+                val = np.argmin(np.abs(limval-spine))
+                if limval > spine.max() or limval < spine.min():
+                    log.warn("The limit {0} is out of bounds."
+                             "  Using min/max instead.".format(lim))
+                if lim[1:] == 'hi':
+                    # End-inclusive indexing: need to add one for the high
+                    # slice
+                    limit_dict[lim] = val + 1
+                else:
+                    limit_dict[lim] = val
+
+        slices = [slice(limit_dict[xx+'lo'], limit_dict[xx+'hi'])
+                  for xx in 'zyx']
+
+        return self[slices]
+
+    def subcube_from_ds9region(self, ds9region):
+        """
+        Extract a masked subcube from a ds9 region or a pyregion Region object
+        (only functions on celestial dimensions)
+
+        Parameters
+        ----------
+        ds9region: str or `pyregion.Shape`
+            The region to extract
+        """
+        import pyregion
+
+        if isinstance(ds9region, six.string_types):
+            shapelist = pyregion.parse(ds9region)
+        else:
+            shapelist = ds9region
+
+        if shapelist[0].coord_format not in ('physical','image'):
+            # Requires astropy >0.4...
+            # pixel_regions = shapelist.as_imagecoord(self.wcs.celestial.to_header())
+            # convert the regions to image (pixel) coordinates
+            celhdr = self.wcs.sub([wcs.WCSSUB_CELESTIAL]).to_header()
+            pixel_regions = shapelist.as_imagecoord(celhdr)
+        else:
+            # For this to work, we'd need to change the reference pixel after cropping.
+            # Alternatively, we can just make the full-sized mask... todo....
+            raise NotImplementedError("Can't use non-celestial coordinates with regions.")
+            pixel_regions = shapelist
+
+        # This is a hack to use mpl to determine the outer bounds of the regions
+        # (but it's a legit hack - pyregion needs a major internal refactor
+        # before we can approach this any other way, I think -AG)
+        mpl_objs = pixel_regions.get_mpl_patches_texts()[0]
+
+        # Find the minimal enclosing box containing all of the regions
+        # (this will speed up the mask creation below)
+        extent = mpl_objs[0].get_extents()
+        xlo, ylo = extent.min
+        xhi, yhi = extent.max
+        all_extents = [obj.get_extents() for obj in mpl_objs]
+        for ext in all_extents:
+            xlo = xlo if xlo < ext.min[0] else ext.min[0]
+            ylo = ylo if ylo < ext.min[1] else ext.min[1]
+            xhi = xhi if xhi > ext.max[0] else ext.max[0]
+            yhi = yhi if yhi > ext.max[1] else ext.max[1]
+
+        log.debug("Region boundaries: ")
+        log.debug("xlo={xlo}, ylo={ylo}, xhi={xhi}, yhi={yhi}".format(xlo=xlo,
+                                                                      ylo=ylo,
+                                                                      xhi=xhi,
+                                                                      yhi=yhi))
+
+        subcube = self.subcube(xlo=xlo, ylo=ylo, xhi=xhi, yhi=yhi)
+        subhdr = subcube.wcs.sub([wcs.WCSSUB_CELESTIAL]).to_header()
+
+        mask = shapelist.get_mask(header=subhdr,
+                                  shape=subcube.shape[1:])
+
+        return subcube.with_mask(BooleanArrayMask(mask, subcube.wcs,
+                                                  shape=subcube.shape))
+
+
 
     def world_spines(self):
         """
@@ -1099,7 +1473,8 @@ class SpectralCube(object):
         cube = read(filename, format=format, hdu=hdu, **kwargs)
         if isinstance(cube, StokesSpectralCube):
             return SpectralCube(data=cube._data, wcs=cube._wcs,
-                                meta=cube._meta, mask=cube._mask)
+                                meta=cube._meta, mask=cube._mask,
+                                header=cube._header)
         else:
             return cube
 
@@ -1119,16 +1494,10 @@ class SpectralCube(object):
         from .io.core import write
         write(filename, self, overwrite=overwrite, format=format)
 
-    def to_yt(self, spectral_factor=1.0, center=None, nprocs=1):
+    def to_yt(self, spectral_factor=1.0, nprocs=None, **kwargs):
         """
-        Convert a spectral cube to a yt object that can be further analyzed in yt.
-
-        By default, the yt object returned will be defined in the default yt
-        spatial units (1 spatial pixel = 1 cm) centered on the center of the
-        spectral cube in all directions. If the ``center`` argument is passed,
-        then the cube is still returned in the default yt spatial units, but
-        shifted so that the specified coordinates are at the origin in the
-        returned object.
+        Convert a spectral cube to a yt object that can be further analyzed in
+        yt.
 
         Parameters
         ----------
@@ -1136,32 +1505,72 @@ class SpectralCube(object):
             Factor by which to stretch the spectral axis. If set to 1, one pixel
             in spectral coordinates is equivalent to one pixel in spatial
             coordinates.
-        center : iterable
-            Tuple or list containing the three coordinates for the center. These
-            should be given as ``(lon, lat, spectral)``.
+
+        If using yt 3.0 or later, additional keyword arguments will be passed
+        onto yt's ``FITSDataset`` constructor. See the yt documentation
+        (http://yt-project.org/docs/3.0/examining/loading_data.html?#fits-data)
+        for details on options for reading FITS data.
         """
 
-        from yt.mods import load_uniform_grid
+        import yt
 
-        data = {'flux': self._get_filled_data(fill=0.)}
+        if ('dev' in yt.__version__ or
+            StrictVersion(yt.__version__) >= StrictVersion('3.0')):
 
-        nz, ny, nx = self.shape
+            from yt.frontends.fits.api import FITSDataset
+            from astropy.io import fits
 
-        dx = nx / 2.
-        dy = ny / 2.
-        dz = nz / 2. * spectral_factor
+            hdu = fits.PrimaryHDU(self._get_filled_data(fill=0.),
+                                  header=self.wcs.to_header())
 
-        # Determine center in pixel coordinates
-        center = self.wcs.wcs_world2pix([center], 0)[0]
+            hdu.header["BUNIT"] = str(self.unit.to_string(format='fits'))
+            hdu.header["BTYPE"] = "flux"
 
-        pf = load_uniform_grid(data, self.shape, 1.,
-                               bbox=np.array([[(-0.5 - center[2]) * spectral_factor, (nz - 0.5 - center[2]) * spectral_factor],
-                                              [-0.5 - center[1], ny - 0.5 - center[1]],
-                                              [-0.5 - center[0], nx - 0.5 - center[0]]]),
-                               nprocs=nprocs, periodicity=(False, False, False))
+            ds = FITSDataset(hdu, nprocs=nprocs,
+                             spectral_factor=spectral_factor, **kwargs)
 
-        return pf
+        else:
 
+            from yt.mods import load_uniform_grid
+
+            data = {'flux': self._get_filled_data(fill=0.).transpose()}
+
+            nz, ny, nx = self.shape
+
+            if nprocs is None:
+                nprocs = 1
+
+            bbox = np.array([[0.5,float(nx)+0.5],
+                             [0.5,float(ny)+0.5],
+                             [0.5,spectral_factor*float(nz)+0.5]])
+
+            ds = load_uniform_grid(data, [nx,ny,nz], 1., bbox=bbox,
+                                   nprocs=nprocs, periodicity=(False, False,
+                                                               False))
+
+        return ytCube(self, ds, spectral_factor=spectral_factor)
+
+    @property
+    def header(self):
+        # Preserve non-WCS information from previous header iteration
+        header = wcs_utils.strip_wcs_from_header(self._header)
+        header.update(self.wcs.to_header())
+        header['BUNIT'] = self.unit.to_string(format='fits')
+        header.insert(2, Card(keyword='NAXIS', value=self._data.ndim))
+        header.insert(3, Card(keyword='NAXIS1', value=self.shape[2]))
+        header.insert(4, Card(keyword='NAXIS2', value=self.shape[1]))
+        header.insert(5, Card(keyword='NAXIS3', value=self.shape[0]))
+        # TODO: incorporate other relevant metadata here
+        return header
+
+    @property
+    def hdu(self):
+        """
+        HDU version of self
+        """
+        from astropy.io import fits
+        hdu = fits.PrimaryHDU(self.filled_data[:].value, header=self.header)
+        return hdu
 
 class StokesSpectralCube(SpectralCube):
 
@@ -1171,14 +1580,16 @@ class StokesSpectralCube(SpectralCube):
     parameters can be accessed with attribute notation.
     """
 
-    def __init__(self, data, wcs, mask=None, meta=None):
+    def __init__(self, data, wcs, mask=None, meta=None, header=None):
 
         # WCS should be 3-d, data should be dict of 3-d, mask should be disk
         # of 3-d
 
         # XXX: For now, let's just extract I and work with that
 
-        super(StokesSpectralCube, self).__init__(data["I"], wcs, mask=mask["I"], meta=meta)
+        super(StokesSpectralCube, self).__init__(data["I"], wcs,
+                                                 mask=mask["I"], meta=meta,
+                                                 header=header)
 
         # TODO: deal with the other stokes parameters here
 
