@@ -3,7 +3,7 @@ import numpy as np
 from astropy.wcs import WCS
 import warnings
 from astropy import units as u
-import logging
+from astropy import log
 
 wcs_parameters_to_preserve = ['cel_offset', 'dateavg', 'dateobs', 'equinox',
                               'latpole', 'lonpole', 'mjdavg', 'mjdobs', 'name',
@@ -160,7 +160,7 @@ def axis_names(wcs):
     return names
 
 
-def slice_wcs(wcs, view):
+def slice_wcs(mywcs, view, numpy_order=True):
     """
     Slice a WCS instance using a Numpy slice. The order of the slice should
     be reversed (as for the data) compared to the natural WCS order.
@@ -168,27 +168,73 @@ def slice_wcs(wcs, view):
     Parameters
     ----------
     view : tuple
-        A tuple containing the same number of slices as the WCS system
+        A tuple containing the same number of slices as the WCS system.
+        The ``step`` method, the third argument to a slice, is not
+        presently supported.
+    numpy_order : bool
+        Use numpy order, i.e. slice the WCS so that an identical slice
+        applied to a numpy array will slice the array and WCS in the same
+        way. If set to `False`, the WCS will be sliced in FITS order,
+        meaning the first slice will be applied to the *last* numpy index
+        but the *first* WCS axis.
 
     Returns
     -------
-    A new `~astropy.wcs.WCS` instance
+    wcs_new : `~astropy.wcs.WCS`
+        A new resampled WCS axis
     """
-    if len(view) != wcs.wcs.naxis:
-        raise ValueError("Must have same number of slices as number of WCS axes")
+    if hasattr(view, '__len__') and len(view) > mywcs.wcs.naxis:
+        raise ValueError("Must have # of slices <= # of WCS axes")
+    elif not hasattr(view, '__len__'): # view MUST be an iterable
+        view = [view]
 
-    wcs_new = wcs.deepcopy()
+    if not all([isinstance(x, slice) for x in view]):
+        raise ValueError("Cannot downsample a WCS with indexing.  Use "
+                         "wcs.sub or wcs.dropaxis if you want to remove "
+                         "axes.")
+
+    wcs_new = mywcs.deepcopy()
     for i, iview in enumerate(view):
+        if iview.step is not None and iview.start is None:
+            # Slice from "None" is equivalent to slice from 0 (but one
+            # might want to downsample, so allow slices with
+            # None,None,step or None,stop,step)
+            iview = slice(0, iview.stop, iview.step)
+
         if iview.start is not None:
+            if numpy_order:
+                wcs_index = mywcs.wcs.naxis - 1 - i
+            else:
+                wcs_index = i
+
             if iview.step not in (None, 1):
-                raise NotImplementedError("Cannot yet slice WCS with strides different from None or 1")
-            wcs_index = wcs.wcs.naxis - 1 - i
-            wcs_new.wcs.crpix[wcs_index] -= iview.start
+                crpix = mywcs.wcs.crpix[wcs_index]
+                cdelt = mywcs.wcs.cdelt[wcs_index]
+                # equivalently (keep this comment so you can compare eqns):
+                # wcs_new.wcs.crpix[wcs_index] =
+                # (crpix - iview.start)*iview.step + 0.5 - iview.step/2.
+                crp = ((crpix - iview.start - 1.)/iview.step
+                       + 0.5 + 1./iview.step/2.)
+                wcs_new.wcs.crpix[wcs_index] = crp
+                wcs_new.wcs.cdelt[wcs_index] = cdelt * iview.step
+            else:
+                wcs_new.wcs.crpix[wcs_index] -= iview.start
     return wcs_new
 
-def check_equality(wcs1, wcs2, warn_missing=False):
+def check_equality(wcs1, wcs2, warn_missing=False, ignore_keywords=['MJD-OBS',
+                                                                    'VELOSYS']):
     """
     Check if two WCSs are equal
+
+    Parameters
+    ----------
+    wcs1, wcs2: `astropy.wcs.WCS`
+        The WCSs
+    warn_missing: bool
+        Issue warnings if one header is missing a keyword that the other has?
+    ignore_keywords: list of str
+        Keywords that are stored as part of the WCS but do not define part of
+        the coordinate system and therefore can be safely ignored.
     """
 
     # naive version:
@@ -213,21 +259,35 @@ def check_equality(wcs1, wcs2, warn_missing=False):
                 u1 = u.Unit(c1[1])
                 u2 = u.Unit(c2[1])
                 if u1 != u2:
-                    OK = False
-                    logging.debug("Header 1, {0}: {1} != {2}".format(key,u1,u2))
+                    if key in ignore_keywords:
+                        log.debug("IGNORED Header 1, {0}: {1} != {2}".format(key,u1,u2))
+                    else:
+                        OK = False
+                        log.debug("Header 1, {0}: {1} != {2}".format(key,u1,u2))
             elif isinstance(c1[1], (float, np.float)):
                 try:
                     np.testing.assert_almost_equal(c1[1], c2[1])
                 except AssertionError:
-                    logging.debug("Header 1, {0}: {1} != {2}".format(key,c1[1],c2[1]))
-                    OK = False
+                    if key in ('RESTFRQ','RESTWAV'):
+                        warnings.warn("{0} is not equal in WCS; ignoring ".format(key)+
+                                      "under the assumption that you want to"
+                                      " compare velocity cubes.")
+                        continue
+                    if key in ignore_keywords:
+                        log.debug("IGNORED Header 1, {0}: {1} != {2}".format(key,c1[1],c2[1]))
+                    else:
+                        log.debug("Header 1, {0}: {1} != {2}".format(key,c1[1],c2[1]))
+                        OK = False
             elif c1[1] != c2[1]:
-                logging.debug("Header 1, {0}: {1} != {2}".format(key,c1[1],c2[1]))
-                OK = False
+                if key in ignore_keywords:
+                    log.debug("IGNORED Header 1, {0}: {1} != {2}".format(key,c1[1],c2[1]))
+                else:
+                    log.debug("Header 1, {0}: {1} != {2}".format(key,c1[1],c2[1]))
+                    OK = False
         else:
             if warn_missing:
                 warnings.warn("WCS2 is missing card {0}".format(key))
-            else:
+            elif key not in ignore_keywords:
                 OK = False
 
     # Check that there aren't any cards in header 2 that were missing from
@@ -241,3 +301,31 @@ def check_equality(wcs1, wcs2, warn_missing=False):
                 OK = False
 
     return OK
+
+def strip_wcs_from_header(header):
+    """
+    Given a header with WCS information, remove ALL WCS information from that
+    header
+    """
+
+    hwcs = WCS(header)
+    wcsh = hwcs.to_header()
+
+    keys_to_keep = [k for k in header
+                    if (k and k not in wcsh and 'NAXIS' not in k)]
+
+    newheader = header.copy()
+    for kw in newheader.keys():
+        if kw not in keys_to_keep:
+            del newheader[kw]
+
+    for kw in ('CRPIX{ii}', 'CRVAL{ii}', 'CDELT{ii}', 'CUNIT{ii}',
+               'CTYPE{ii}', 'PC0{ii}_0{jj}', 'CD{ii}_{jj}',):
+        for ii in range(5):
+            for jj in range(5):
+                k = kw.format(ii=ii,jj=jj)
+                if k in newheader.keys():
+                    del newheader[k]
+
+
+    return newheader
